/////////////////////////////////////////
//      Nebula M42 by Polecatsoft      //
/////////////////////////////////////////
/* Description:                        
   These class can be used to read and write data
   from and to *.ini files.
///////////////////////////////////////*/

#include <fstream>
#include "iniReader.h"
#include "main.h"
using namespace std;

//Constructor sets category to -1 (-1 is at global scope)
iniReader::iniReader()
{
	curCategory = -1;
	argv = 0;
	curEntry = entries.end();
}

iniReader::~iniReader()
{
	cleanUpArgv();
}

//Clean up the memory returned by getArgvValue()
void iniReader::cleanUpArgv()
{
	if(argv)
		delete [] argv;
	argv = 0;
}

//This will load all the values from the file filename into memory
bool iniReader::load(const string filename)
{
	ifstream ini(filename.c_str());
	if(ini.fail())
		return false;
	string buf;
	int category = -1;
	while(!ini.eof())
	{
		getline(ini, buf, '\n');
		for(int i = 0; i < buf.length(); ++i)
		{
			if(buf[i] <= ' ')
				continue;
			if(buf[i] == ';')
				break;
			if(buf[i] == '[')
			{
				string temp;
				while(++i < buf.length() && buf[i] != ']')
					temp.append(1, buf[i]);
				categories.push_back(temp);
				++category;
				continue;
			}
			iniEntry newEntry;
			newEntry.category = category;
			while(buf[i] != '=' && buf[i] > ' ')
			{
				newEntry.name.append(1, buf[i++]);
				if(i == buf.length())
					break;
			}
			while(buf[i] <= ' ') ++i;
			if(buf[i++] != '=')
			{
				entries.push_back(newEntry);
				//Decrement i again, so the first character can be used
				i -= 2;
				continue;
			}
			while(buf[i] <= ' ') ++i;
			newEntry.value = buf.erase(0, i);
			entries.push_back(newEntry);
			//allegro_message("New entry: %s=%s, category=%i", newEntry.name.c_str(), newEntry.value.c_str(), newEntry.category);
			break;
		}
	}
	ini.close();
	firstCategoryEntry.reserve(categories.size());
	for(vector<iniEntry>::iterator i = entries.begin(); i != entries.end(); ++i)
	{
		if(i == entries.begin())
		{
			if((*i).category != -1)
				firstCategoryEntry.push_back(i);
		}
		vector<iniEntry>::iterator next = i;
		++next;
		if((*i).category != (*next).category)
			firstCategoryEntry.push_back(next);
	}
	return true;
}

//This will set the category to newCategory, or to the global category if
//newCategory is not found
int iniReader::setCategory(const string newCategory)
{
	for(int i = 0; i < categories.size(); ++i)
	{
		if(categories[i] == newCategory)
		{
			curCategory = i;
			return curCategory;
		}
	}
	curCategory = -1;
	return curCategory;
}

//Returns the value of name in the current category or defaultValue if it's
//nonexistant
string iniReader::getValue(const std::string name, const std::string defaultValue)
{
	vector<iniEntry>::iterator i;
	if(curCategory == -1)
		i = entries.begin();
	else
		i = firstCategoryEntry[curCategory];
	while((*i).category == curCategory)
	{
		if((*i).name == name)
		{
			return (*i).value;
		}
		++i;
	}
	return defaultValue;
}

//Returns an integer version of the value or defaultValue if it's nonexistant
int iniReader::getIValue(const std::string name, const int defaultValue)
{
	return str2i(getValue(name), defaultValue);
}

//Returns a floating point version of the value or defaultValue if it's nonexistant
float iniReader::getFValue(const std::string name, const float defaultValue)
{
	return str2f(getValue(name), defaultValue);
}

//Returns a boolean version of the value or defaultValue if it's nonexistant
bool iniReader::getBValue(const std::string name, const bool defaultValue)
{
	string value = getValue(name);
	if(value == "0" || value == "false")
		return false;
	if(value == "1" || value == "true")
		return true;
	return defaultValue;
}

//Returns an array of strings. It extracts values from the value the entry
//name, that are seperated with seperator. values is set to the number of
//entries returned.
//Returns 0 on error.
std::string *iniReader::getArgvValue(const std::string name, int &values,
	const char seperator)
{
	cleanUpArgv();
	string temp = getValue(name);
	if(temp.empty())
	{
		values = 0;
		return argv;
	}

	values = 1;
	for(int i = 0; i < temp.length(); ++i)
		if(temp[i] == seperator)
			++values;
	argv = new string[values];
	for(int i = 0, iCurEntry = 0; i < temp.length(); ++i)
	{
		if(temp[i] == seperator)
			++iCurEntry;
		else
			argv[i].append(1, temp[i]);
	}

	return argv;
}

//Sets the global category as the current category
void iniReader::setGlobalCategory()
{
	curCategory = -1;
}

//Returns the number of entries loaded
int iniReader::getNumberOfEntries()
{
	return entries.size();
}

//Returns the name of the current category
//Returns an empty string if urcategory is global
std::string iniReader::getCurCategory()
{
	if(curCategory == -1)
		return "";
	return categories[curCategory];
}

//Returns the name of the next category
//Returns an empty string if no further categories exist
std::string iniReader::getNextCategory()
{
	if(curCategory + 1 == categories.size())
		return "";
	return categories[curCategory + 1];
}

//Resets the curEntry to the first entry of the current category
void iniReader::resetEntryPointer()
{
	if(curCategory == -1)
		curEntry = entries.begin();
	else
		curEntry = firstCategoryEntry[curCategory];
}

//Returns the next iniEntry of the current category or 0 on error
iniEntry *iniReader::getNextEntry()
{
	if(curEntry == entries.end())
		return 0;
	iniEntry *retVal = &(*curEntry);
	if(retVal->category != curCategory)
		return 0;
	++curEntry;
	return retVal;
}

//Constructor sets category to -1 (-1 is at global scope)
iniWriter::iniWriter()
{
	curCategory = -1;
}

//This will set the category to newCategory, or to the global category if
//newCategory is not found
void iniWriter::setCategory(const string newCategory)
{
	for(int i = 0; i < categories.size(); ++i)
	{
		if(categories[i] == newCategory)
		{
			curCategory = i;
			return;
		}
	}
	curCategory = -1;
}

//Add a category to the list
void iniWriter::addCategory(const string name)
{
	int found = -1;
	for(int i = 0; i < categories.size(); ++i)
	{
		if(categories[i] == name)
		{
			found = i;
			break;
		}
	}
	if(found == -1)
	{
		categories.push_back(name);
		curCategory = categories.size() - 1;
	}
	else
	{
		curCategory = found;
	}
}

//Sets the global category as the current category
void iniWriter::setGlobalCategory()
{
	curCategory = -1;
}

//Adds a value to the active category
void iniWriter::addEntry(const std::string name, const std::string value)
{
	iniEntry temp;
	temp.name = name;
	temp.value = value;
	temp.category = curCategory;
	entries.push_back(temp);
	//allegro_message("New entry: %s=%s, cat=%i", name.c_str(), value.c_str(), curCategory);
}

void iniWriter::addEntry(const std::string name, const int value)
{
	addEntry(name, i2str(value));
}

void iniWriter::addEntry(const std::string name, const float value)
{
	addEntry(name, f2str(value));
}

void iniWriter::addEntry(const std::string name, const bool value)
{
	addEntry(name, value ? string("true") : string("false"));
}

//This will save all the values from memory into the file filename
void iniWriter::save(const std::string filename)
{
	ofstream file(filename.c_str());
	for(int i = -1; i < static_cast<int>(categories.size()); ++i)
	{
		if(i > -1)
		{
			file << "[" << categories[i] << "]" << endl;
		}
		for(int y = 0; y < entries.size(); ++y)
		{
			if(entries[y].category == i)
				file << entries[y].name << "=" << entries[y].value << endl;
		}
		if(i + 1 != static_cast<int>(categories.size()))
			file << endl;
	}
	file.close();
}

//Adds all entries from reader to this iniWriter instance
void iniWriter::copyAll(iniReader &reader)
{
	reader.setGlobalCategory();
	string category;
	do
	{
		if(category != "")
		{
			addCategory(category);
		}
		else
		{
			setGlobalCategory();
		}
		reader.resetEntryPointer();
		iniEntry *curEntry;
		while((curEntry = reader.getNextEntry()) != 0)
		{
			addEntry(curEntry->name, curEntry->value);
		}
	} while((category = reader.getNextCategory()) != "");
}
